@@ -151,7 +151,6 @@ gboolean   meta_core_begin_grab_op (Display    *xdisplay,
                                     MetaGrabOp  op,
                                     gboolean    pointer_already_grabbed,
                                     gboolean    frame_action,
-                                    int         event_serial,
                                     int         button,
                                     gulong      modmask,
                                     guint32     timestamp,
@@ -182,8 +181,6 @@ void       meta_core_get_screen_size (Display *xdisplay,
  */
 void meta_core_increment_event_serial (Display *display);
 
-int meta_ui_get_last_event_serial (Display *xdisplay);
-
 void meta_invalidate_default_icons (void);
 
 #endif
