@@ -33,6 +33,8 @@
 #include "fixedtip.h"
 #include "theme.h"
 #include "prefs.h"
+#include "errors.h"
+#include "ui.h"
 
 #ifdef HAVE_SHAPE
 #include <X11/extensions/shape.h>
@@ -479,6 +481,8 @@ meta_frames_ensure_layout (MetaFrames  *frames,
                                           flags);
       
       frame->layout = gtk_widget_create_pango_layout (widget, frame->title);
+
+      pango_layout_set_auto_dir (frame->layout, FALSE);
       
       font_desc = meta_gtk_widget_get_font_desc (widget, scale,
 						 meta_prefs_get_titlebar_font ());
@@ -570,9 +574,6 @@ meta_frames_manage_window (MetaFrames *frames,
 
   gdk_window_set_user_data (frame->window, frames);
 
-  /* Set the window background to the current style */
-  meta_frames_set_window_background (frames, frame);
-
   /* Don't set event mask here, it's in frame.c */
   
   frame->xwindow = xwindow;
@@ -583,6 +584,11 @@ meta_frames_manage_window (MetaFrames *frames,
   frame->expose_delayed = FALSE;
   frame->shape_applied = FALSE;
   frame->prelit_control = META_FRAME_CONTROL_NONE;
+
+  /* Don't set the window background yet; we need frame->xwindow to be
+   * registered with its MetaWindow, which happens after this function
+   * and meta_ui_create_frame_window() return to meta_window_ensure_frame().
+   */
   
   meta_core_grab_buttons (gdk_display, frame->xwindow);
   
@@ -788,12 +794,13 @@ meta_frames_apply_shapes (MetaFrames *frames,
   
   if (fgeom.top_left_corner_rounded_radius != 0)
     {
-      const int radius = fgeom.top_left_corner_rounded_radius;
+      const int corner = fgeom.top_left_corner_rounded_radius;
+      const float radius = sqrt(corner) + corner;
       int i;
 
-      for (i=0; i<radius; i++)
+      for (i=0; i<corner; i++)
         {
-          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i)) + 0.5));
+          const int width = floor(0.5 + radius - sqrt(radius*radius - (radius-(i+0.5))*(radius-(i+0.5))));
           xrect.x = 0;
           xrect.y = i;
           xrect.width = width;
@@ -805,12 +812,13 @@ meta_frames_apply_shapes (MetaFrames *frames,
 
   if (fgeom.top_right_corner_rounded_radius != 0)
     {
-      const int radius = fgeom.top_right_corner_rounded_radius;
+      const int corner = fgeom.top_right_corner_rounded_radius;
+      const float radius = sqrt(corner) + corner;
       int i;
 
-      for (i=0; i<radius; i++)
+      for (i=0; i<corner; i++)
         {
-          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i)) + 0.5));
+          const int width = floor(0.5 + radius - sqrt(radius*radius - (radius-(i+0.5))*(radius-(i+0.5))));
           xrect.x = new_window_width - width;
           xrect.y = i;
           xrect.width = width;
@@ -822,14 +830,15 @@ meta_frames_apply_shapes (MetaFrames *frames,
 
   if (fgeom.bottom_left_corner_rounded_radius != 0)
     {
-      const int radius = fgeom.bottom_left_corner_rounded_radius;
+      const int corner = fgeom.bottom_left_corner_rounded_radius;
+      const float radius = sqrt(corner) + corner;
       int i;
 
-      for (i=0; i<radius; i++)
+      for (i=0; i<corner; i++)
         {
-          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i)) + 0.5));
+          const int width = floor(0.5 + radius - sqrt(radius*radius - (radius-(i+0.5))*(radius-(i+0.5))));
           xrect.x = 0;
-          xrect.y = new_window_height - i;
+          xrect.y = new_window_height - i - 1;
           xrect.width = width;
           xrect.height = 1;
           
@@ -839,14 +848,15 @@ meta_frames_apply_shapes (MetaFrames *frames,
 
   if (fgeom.bottom_right_corner_rounded_radius != 0)
     {
-      const int radius = fgeom.bottom_right_corner_rounded_radius;
+      const int corner = fgeom.bottom_right_corner_rounded_radius;
+      const float radius = sqrt(corner) + corner;
       int i;
 
-      for (i=0; i<radius; i++)
+      for (i=0; i<corner; i++)
         {
-          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i)) + 0.5));
+          const int width = floor(0.5 + radius - sqrt(radius*radius - (radius-(i+0.5))*(radius-(i+0.5))));
           xrect.x = new_window_width - width;
-          xrect.y = new_window_height - i;
+          xrect.y = new_window_height - i - 1;
           xrect.width = width;
           xrect.height = 1;
           
@@ -1127,6 +1137,11 @@ show_tip_now (MetaFrames *frames)
       /* get conversion delta for root-to-frame coords */
       dx = root_x - x;
       dy = root_y - y;
+
+      /* Align the tooltip to the button right end if RTL */
+      if (meta_ui_get_direction() == META_UI_DIRECTION_RTL)
+        dx += rect->width;
+
       screen_number = gdk_screen_get_number (gtk_widget_get_screen (GTK_WIDGET (frames)));
       
       meta_fixed_tip_show (gdk_display,
@@ -1187,6 +1202,108 @@ redraw_control (MetaFrames *frames,
   invalidate_cache (frames, frame);
 }
 
+static gboolean
+meta_frame_titlebar_event (MetaUIFrame    *frame,
+                           GdkEventButton *event,
+                           int            action)
+{
+  MetaFrameFlags flags;
+  
+  switch (action)
+    {
+    case META_ACTION_TITLEBAR_TOGGLE_SHADE:
+      {
+        flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
+        
+        if (flags & META_FRAME_ALLOWS_SHADE)
+          {
+            if (flags & META_FRAME_SHADED)
+              meta_core_unshade (gdk_display,
+                                 frame->xwindow,
+                                 event->time);
+            else
+              meta_core_shade (gdk_display,
+                               frame->xwindow,
+                               event->time);
+          }
+      }
+      break;          
+      
+    case META_ACTION_TITLEBAR_TOGGLE_MAXIMIZE:
+      {
+        flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
+        
+        if (flags & META_FRAME_ALLOWS_MAXIMIZE)
+          {
+            meta_core_toggle_maximize (gdk_display, frame->xwindow);
+          }
+      }
+      break;
+
+    case META_ACTION_TITLEBAR_MINIMIZE:
+      {
+        flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
+        
+        if (flags & META_FRAME_ALLOWS_MINIMIZE)
+          {
+            meta_core_minimize (gdk_display, frame->xwindow);
+          }
+      }
+      break;
+
+    case META_ACTION_TITLEBAR_NONE:
+      /* Yaay, a sane user that doesn't use that other weird crap! */
+      break;
+    
+    case META_ACTION_TITLEBAR_LOWER:
+      meta_core_user_lower_and_unfocus (gdk_display,
+                                        frame->xwindow,
+                                        event->time);
+      break;
+
+    case META_ACTION_TITLEBAR_MENU:
+      meta_core_show_window_menu (gdk_display,
+                                  frame->xwindow,
+                                  event->x_root,
+                                  event->y_root,
+                                  event->button,
+                                  event->time);
+      break;
+
+    case META_ACTION_TITLEBAR_LAST:
+      break;
+    }
+  
+  return TRUE;
+}
+
+static gboolean
+meta_frame_double_click_event (MetaUIFrame    *frame,
+                               GdkEventButton *event)
+{
+  int action = meta_prefs_get_action_double_click_titlebar ();
+  
+  return meta_frame_titlebar_event (frame, event, action);
+}
+
+static gboolean
+meta_frame_middle_click_event (MetaUIFrame    *frame,
+                               GdkEventButton *event)
+{
+  int action = meta_prefs_get_action_middle_click_titlebar();
+  
+  return meta_frame_titlebar_event (frame, event, action);
+}
+
+static gboolean
+meta_frame_right_click_event(MetaUIFrame     *frame,
+                             GdkEventButton  *event)
+{
+  int action = meta_prefs_get_action_right_click_titlebar();
+  
+  return meta_frame_titlebar_event (frame, event, action);
+}
+
 static gboolean
 meta_frames_button_press_event (GtkWidget      *widget,
                                 GdkEventButton *event)
@@ -1233,59 +1350,7 @@ meta_frames_button_press_event (GtkWidget      *widget,
       event->button == 1 &&
       event->type == GDK_2BUTTON_PRESS)
     {
-      MetaFrameFlags flags;
-      
-      switch (meta_prefs_get_action_double_click_titlebar ())
-        {
-        case META_ACTION_DOUBLE_CLICK_TITLEBAR_TOGGLE_SHADE:
-          {
-            flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
-            
-            if (flags & META_FRAME_ALLOWS_SHADE)
-              {
-                if (flags & META_FRAME_SHADED)
-                  meta_core_unshade (gdk_display,
-                                     frame->xwindow,
-                                     event->time);
-                else
-                  meta_core_shade (gdk_display,
-                                   frame->xwindow,
-                                   event->time);
-              }
-          }
-          break;          
-          
-        case META_ACTION_DOUBLE_CLICK_TITLEBAR_TOGGLE_MAXIMIZE:
-          {
-            flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
-            
-            if (flags & META_FRAME_ALLOWS_MAXIMIZE)
-              {
-                meta_core_toggle_maximize (gdk_display, frame->xwindow);
-              }
-          }
-          break;
-
-        case META_ACTION_DOUBLE_CLICK_TITLEBAR_MINIMIZE:
-          {
-            flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
-            
-            if (flags & META_FRAME_ALLOWS_MINIMIZE)
-              {
-                meta_core_minimize (gdk_display, frame->xwindow);
-              }
-          }
-          break;
-
-        case META_ACTION_DOUBLE_CLICK_TITLEBAR_NONE:
-          /* Yaay, a sane user that doesn't use that other weird crap! */
-          break;
-
-        case META_ACTION_DOUBLE_CLICK_TITLEBAR_LAST:
-          break;
-        }
-      
-      return TRUE;
+      return meta_frame_double_click_event (frame, event);
     }
 
   if (meta_core_get_grab_op (gdk_display) !=
@@ -1352,7 +1417,6 @@ meta_frames_button_press_event (GtkWidget      *widget,
                                op,
                                TRUE,
                                TRUE,
-                               meta_ui_get_last_event_serial (gdk_display),
                                event->button,
                                0,
                                event->time,
@@ -1376,6 +1440,10 @@ meta_frames_button_press_event (GtkWidget      *widget,
           dx = event->x_root - event->x;
           dy = event->y_root - event->y;
           
+	  /* Align to the right end of the menu rectangle if RTL */
+	  if (meta_ui_get_direction() == META_UI_DIRECTION_RTL)
+	    dx += rect->width;
+
           meta_core_show_window_menu (gdk_display,
                                       frame->xwindow,
                                       rect->x + dx,
@@ -1443,7 +1511,6 @@ meta_frames_button_press_event (GtkWidget      *widget,
                                  op,
                                  TRUE,
                                  TRUE,
-                                 meta_ui_get_last_event_serial (gdk_display),
                                  event->button,
                                  0,
                                  event->time,
@@ -1464,7 +1531,6 @@ meta_frames_button_press_event (GtkWidget      *widget,
                                    META_GRAB_OP_MOVING,
                                    TRUE,
                                    TRUE,
-                                   meta_ui_get_last_event_serial (gdk_display),
                                    event->button,
                                    0,
                                    event->time,
@@ -1474,18 +1540,11 @@ meta_frames_button_press_event (GtkWidget      *widget,
     }
   else if (event->button == 2)
     {
-      meta_core_user_lower_and_unfocus (gdk_display,
-                                        frame->xwindow,
-                                        event->time);
+      return meta_frame_middle_click_event (frame, event);
     }
   else if (event->button == 3)
     {
-      meta_core_show_window_menu (gdk_display,
-                                  frame->xwindow,
-                                  event->x_root,
-                                  event->y_root,
-                                  event->button,
-                                  event->time);
+      return meta_frame_right_click_event (frame, event);
     }
   
   return TRUE;
@@ -1797,7 +1856,8 @@ meta_frames_motion_notify_event     (GtkWidget           *widget,
                 grab_op == META_GRAB_OP_CLICKING_DELETE) ||
                (control == META_FRAME_CONTROL_MINIMIZE &&
                 grab_op == META_GRAB_OP_CLICKING_MINIMIZE) ||
-               (control == META_FRAME_CONTROL_MAXIMIZE &&
+               ((control == META_FRAME_CONTROL_MAXIMIZE ||
+                 control == META_FRAME_CONTROL_UNMAXIMIZE) &&
                 (grab_op == META_GRAB_OP_CLICKING_MAXIMIZE ||
                  grab_op == META_GRAB_OP_CLICKING_UNMAXIMIZE)) ||
                (control == META_FRAME_CONTROL_SHADE &&
@@ -2129,6 +2189,11 @@ meta_frames_expose_event (GtkWidget           *widget,
   return TRUE;
 }
 
+/* How far off the screen edge the window decorations should
+ * be drawn. Used only in meta_frames_paint_to_drawable, below.
+ */
+#define DECORATING_BORDER 100
+
 static void
 meta_frames_paint_to_drawable (MetaFrames   *frames,
                                MetaUIFrame  *frame,
@@ -2146,12 +2211,6 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
   MetaButtonState button_states[META_BUTTON_TYPE_LAST];
   Window grab_frame;
   int i;
-  int top, bottom, left, right;
-  GdkRegion *edges;
-  GdkRegion *tmp_region;
-  GdkRectangle area;
-  GdkRectangle *areas;
-  int n_areas;
   MetaButtonLayout button_layout;
   MetaGrabOp grab_op;
   
@@ -2262,35 +2321,95 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
 
   meta_frames_ensure_layout (frames, frame);
 
-  meta_theme_get_frame_borders (meta_theme_get_current (),
-                                type, frame->text_height, flags, 
-                                &top, &bottom, &left, &right);
-
-  /* Repaint each side of the frame */
-  
-  edges = gdk_region_copy (region);
-
-  /* Punch out the client area */
-  area.x = left;
-  area.y = top;
-  area.width = w;
-  area.height = h;
-  tmp_region = gdk_region_rectangle (&area);
-  gdk_region_subtract (edges, tmp_region);
-  gdk_region_destroy (tmp_region);
-
-  /* Now draw remaining portion of region */
-  gdk_region_get_rectangles (edges, &areas, &n_areas);
-
   meta_prefs_get_button_layout (&button_layout);
-  for (i = 0; i < n_areas; i++)
+
+  if (G_LIKELY (GDK_IS_WINDOW (drawable)))
     {
-      if (GDK_IS_WINDOW (drawable))
-        gdk_window_begin_paint_rect (drawable, &areas[i]);
+      /* A window; happens about 2/3 of the time */
+
+      GdkRectangle area, *areas;
+      int n_areas;
+      int screen_width, screen_height;
+      GdkRegion *edges, *tmp_region;
+      int top, bottom, left, right;
+ 
+      /* Repaint each side of the frame */
+
+      meta_theme_get_frame_borders (meta_theme_get_current (),
+                             type, frame->text_height, flags, 
+                             &top, &bottom, &left, &right);
+
+      meta_core_get_screen_size (gdk_display,
+                             frame->xwindow,
+                             &screen_width, &screen_height);
+
+      edges = gdk_region_copy (region);
+
+      /* Punch out the client area */
+
+      area.x = left;
+      area.y = top;
+      area.width = w;
+      area.height = h;
+      tmp_region = gdk_region_rectangle (&area);
+      gdk_region_subtract (edges, tmp_region);
+      gdk_region_destroy (tmp_region);
+
+      /* Now draw remaining portion of region */
+
+      gdk_region_get_rectangles (edges, &areas, &n_areas);
+
+      for (i = 0; i < n_areas; i++)
+        {
+          /* Bug 399529: clamp areas[i] so that it doesn't go too far
+           * off the edge of the screen. This works around a GDK bug
+           * which makes gdk_window_begin_paint_rect cause an X error
+           * if the window is insanely huge. If the client is a GDK program
+           * and does this, it will still probably cause an X error in that
+           * program, but the last thing we want is for Metacity to crash
+           * because it attempted to decorate the silly window.
+           */
+
+          areas[i].x = MAX (areas[i].x, -DECORATING_BORDER); 
+          areas[i].y = MAX (areas[i].y, -DECORATING_BORDER); 
+          if (areas[i].x+areas[i].width  > screen_width  + DECORATING_BORDER)
+            areas[i].width  = MIN (0, screen_width  - areas[i].x);
+          if (areas[i].y+areas[i].height > screen_height + DECORATING_BORDER)
+            areas[i].height = MIN (0, screen_height - areas[i].y);
+
+          /* Okay, so let's start painting. */
+
+          gdk_window_begin_paint_rect (drawable, &areas[i]);
+
+          meta_theme_draw_frame (meta_theme_get_current (),
+            widget,
+            drawable,
+            NULL, /* &areas[i], */
+            x_offset, y_offset,
+            type,
+            flags,
+            w, h,
+            frame->layout,
+            frame->text_height,
+            &button_layout,
+            button_states,
+            mini_icon, icon);
+
+          gdk_window_end_paint (drawable);
+        }
+
+      g_free (areas);
+      gdk_region_destroy (edges);
+
+    }
+  else
+    {
+      /* Not a window; happens about 1/3 of the time */
+
       meta_theme_draw_frame (meta_theme_get_current (),
                              widget,
                              drawable,
-                             NULL, /* &areas[i], */
+                             NULL,
                              x_offset, y_offset,
                              type,
                              flags,
@@ -2300,13 +2419,8 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
                              &button_layout,
                              button_states,
                              mini_icon, icon);
-
-      if (GDK_IS_WINDOW (drawable))
-        gdk_window_end_paint (drawable);
     }
 
-  gdk_region_destroy (edges);
-  g_free (areas);
 }
 
 static void
